% Code (c) Charis Lanaras, ETH Zurich, Oct 28 2015 (see LICENSE)
% charis.lanaras@geod.baug.ethz.ch

function [D_RGB, D_XRF, Dh_XRF, A, AS, Ah, ASh, stat] = dqq_SupResPALM_nv_GD_TV_Dh_adv_Square_S(XRF_l_signal, RGB_signal, D_XRF_ini, Dh_XRF_ini, D_RGB_ini, supParam )
% SupResPALM - Perform RGB-XRF super-resolution by fusion

% Usage
%   [D_RGB, D_XRF, Dh_XRF, A, AS, Ah, ASh, stat] = dqq_SupResPALM_nv_GD_TV_Dh_adv_Square_S(XRF_l_signal, RGB_signal, D_XRF_ini, Dh_XRF_ini, D_RGB_ini, supParam )
% 
% Inputs
%   XRF_l_signal  - LR XRF image (in 2D format)
%   RGB_signal    - HR RGB image (in 2D format)
%   D_XRF_ini     - Initialization for XRF dictionary, visible component
%   Dh_XRF_ini    - Initialization for XRF dictionary, non-visible component
%   D_RGB_ini     - Initialization for RGB dictionary
%   supParam      - Parameters to the algorithm
%
% Outputs
%   D_RGB         - Output RGB dictionary
%   D_XRF         - Output XRF dictionary, visible component
%   Dh_XRF        - Output XRF dictionary, non-visible component
%   A             - Output coefficients in HR, visible component
%   AS            - Output coefficients in LR, visible component, (AS = A * S)
%   Ah            - Output coefficients in HR, non-visible component
%   ASh           - Output coefficients in LR, non-visible component, (AS = A * S)
%   stat          - Stats for the optimization
%
% References
% Dai, Qiqin, Emeline Pouyet, Oliver Cossairt, Marc Walton, and Aggelos K. Katsaggelos. "Spatial-Spectral Representation for X-Ray Fluorescence Image Super-Resolution." 
% IEEE Transactions on Computational Imaging (2017).
% 
% Qiqin Dai
% 10/22/2017

%   Get the parameters from input structure

maxIter = supParam.maxIter;
epsilon = supParam.epsilon;
p = supParam.p;
TVlambda = supParam.TVlambda;
s2=p/2; %Active endmemebers per pixel (on average over the whole image)
param = supParam.param;


resTotal(1) = 1000;



im_RGB = hyperConvert3d(RGB_signal, supParam.height, supParam.width);

scale_diff = supParam.scale_diff;

%   Generate the degradation matrix
S = hyperSpatialDown(supParam.height, supParam.width, scale_diff);

D_XRF  = D_XRF_ini;
Dh_XRF = Dh_XRF_ini;
D_RGB  = D_RGB_ini;


AS = mexLasso(XRF_l_signal, D_XRF, param);
AS = full(AS);
A = hyperConvert2d(imresize(hyperConvert3d(AS,supParam.height/scale_diff, supParam.width/scale_diff),scale_diff));
Ah = zeros(size(A));
ASh = zeros(size(AS));

AS = sparse(AS);
ASh = sparse(ASh);



%   Prepare the TV matrix M

% M = sparse((size(A, 2)), 2 * (sqrt(size(A,2)) - 1) * (sqrt(size(A,2)) - 1));
% im_RGB_size = size(im_RGB);
% index_size = [im_RGB_size(1), im_RGB_size(2)];
% 
% kk = 1;
%
% for i = 1 : im_RGB_size(1) - 1
% 
%     for j = 1 : im_RGB_size(2) - 1
%         
% 
% 
%         index_cur = sub2ind(index_size, i, j);  
%         index_cur_right = sub2ind(index_size, i, j + 1);
%         index_cur_down = sub2ind(index_size, i + 1, j );
%         
%         M(index_cur, kk) = 1;
%         M(index_cur_right,kk) = -1;
%         kk = kk + 1;
% 
%         M(index_cur, kk) = 1;
%         M(index_cur_down,kk) = -1;
%         kk = kk + 1;
%         
%     end
% 
% 
% end

%   Load the precomputed Mask_M to save time  
load('Mask_M.mat');


for j=1:maxIter
    
    A_old = A;
    Ah_old = Ah;
    D_RGB_old = D_RGB;
    D_XRF_old = D_XRF;
    Dh_XRF_old = Dh_XRF;
    
    
    %   Update A and Ah simultaneously, using TV regularization
    [ ~, A, ~, Ah, res_1 ] = highResStepAAh( RGB_signal, D_RGB, A, XRF_l_signal, Dh_XRF, Ah, D_XRF, S, M, TVlambda, s2 );


    resRGB_A(j) = norm(RGB_signal - D_RGB * A, 'fro');

    %   Update AS
    AS = A*S;
    %   Update ASh
    ASh = Ah*S;
    
    %   Update ASh
    resASh(j) = norm(XRF_l_signal - D_XRF * AS - Dh_XRF * ASh, 'fro');

    resTV(j) = norm(Dh_XRF * Ah * M, 'fro');
    
    %   Update D_RGB
    [D_RGB, ~, res_2] = lowResStep(RGB_signal, D_RGB, A);
    resD_RGB(j) = norm(RGB_signal - D_RGB * A, 'fro');
    
    %   Update D_XRF
    [D_XRF, ~, res_3] = lowResStep((XRF_l_signal - Dh_XRF * ASh), D_XRF, AS);
    resD_XRF(j) = norm(XRF_l_signal - D_XRF * AS - Dh_XRF * ASh, 'fro');

    %   Update Dh_XRF
    [Dh_XRF, ~, res_4] = lowResStep_HR_TV((XRF_l_signal - D_XRF * AS), Dh_XRF, ASh, Ah, M, TVlambda);
    resDh_XRF(j) = norm(XRF_l_signal - D_XRF * AS - Dh_XRF * ASh, 'fro');
    resTV_2(j) = norm(Dh_XRF * Ah * M, 'fro');
    
    %   Calculate statics
    resTotal(j+1) = resD_RGB(j)+resDh_XRF(j) + TVlambda * resTV_2(j);
    
    
    fprintf('%6s %16s %16s %16s %16s  %16s %16s %16s %16s\n', 'Iter', 'total_res', 'resRGB_A', 'resASh', 'resTV', 'resD_RGB', 'resD_XRF', 'resDh_XRF', 'resTV' );
    fprintf('%6i %16.2e %16.2e %16.2e %16.2e %16.2e %16.2e %16.2e %16.2e\n', j, resTotal(j+1), resRGB_A(j), resASh(j), resTV(j), resD_RGB(j), resD_XRF(j), resDh_XRF(j), resTV_2(j) ); 


    if (resTotal(j+1) / resTotal(j)) > 1 || (abs(resTotal(j+1) - resTotal(j)) / resTotal(j) ) < epsilon 
        A = A_old;
        Ah = Ah_old;
        D_RGB = D_RGB_old;
        D_XRF = D_XRF_old;
        Dh_XRF = Dh_XRF_old;
        break
    end 
    
    
end


%   Gather the stat
stat = [];
stat.resTotal = resTotal;
stat.resRGB_A = resRGB_A;
stat.resRGB_AS = resD_RGB;
stat.resD_XRF = resD_XRF;
stat.resASh = resASh;
stat.resDh_XRF = resDh_XRF;
stat.resTV = resTV;

end




function [ D_RGB, A,  Dh_XRF, Ah, res ] = highResStepAAh( RGB_l_signal, D_RGB, A, XRF_l_signal, Dh_XRF, Ah, D_XRF, S, M, TVlambda, sparse_factor )

maxIter = 100;
epsilon = 1.01;
gamma2 = 1.01;

N = size(RGB_l_signal, 2);
beta = round(sparse_factor*N); % Number of desirable non-zero entries

res(1) = norm(RGB_l_signal-D_RGB*A,'fro')+ norm(XRF_l_signal - D_XRF * A * S - Dh_XRF * Ah * S,'fro') + TVlambda * norm(Dh_XRF * Ah * M, 'fro') + 100;

for k=1:maxIter
    D_RGB_old = D_RGB;
    A_old = A;

    Dh_XRF_old = Dh_XRF;
    Ah_old = Ah;
    
    %  Update the Abundances
    dk = gamma2 * norm( D_RGB*D_RGB' ,'fro');
    V = A - 1/dk * D_RGB' * ( D_RGB * A - RGB_l_signal );

    VS = V * S;
    
    dkh = gamma2 * norm( Dh_XRF*Dh_XRF' ,'fro');
    Vh = Ah - 1/dkh * ( Dh_XRF' * ( Dh_XRF * Ah * S - (XRF_l_signal - D_XRF * VS ) ) * S' + TVlambda * Dh_XRF' * (Dh_XRF * Ah * M) * M' );
    

    [A, Ah] = PplusbAAh(V, Vh);
    
    VSh = Vh * S;

    
    % Calculation of residuals

    res(k+1,1) = norm(RGB_l_signal-D_RGB*A,'fro')+ norm(XRF_l_signal - D_XRF * A * S - Dh_XRF * Ah * S,'fro') + TVlambda * norm(Dh_XRF * Ah * M, 'fro');

    % Checks for exiting iteration
    if (1/res(k+1) * res(k)) < epsilon
        break
    end
    
    if (res(k+1) / res(k))>1
        D_RGB = D_RGB_old;
        A = A_old;
        Ah = Ah_old;
        break
    end
    
end
end



function [ E, Ah, res ] = lowResStep_HR_TV( H, E, ASh, Ah, M, TVlambda )

maxIter = 100;
epsilon = 1.01;
gamma1 = 1.01;

notfirst = 0;

res(1) = norm(H-E*ASh,'fro') + TVlambda * norm(E * Ah * M, 'fro') + 100;

for k=1:maxIter
    E_old = E;
    Ah_old = Ah;
    ASh_old = ASh;
    
    % 2.1. Update of signatures
    ck = gamma1 * norm(Ah*Ah','fro');
    U = E - 1/ck * (( E*ASh - H ) * ASh' + TVlambda * E * Ah * (M * M') * Ah');
    E = Pplusa(U);
    
    % Calculation of residuals
    res(k+1,1) = norm(H-E*ASh,'fro') + TVlambda * norm(E * Ah * M, 'fro');
    
    % Checks for exiting iteration
    if (1/res(k+1) * res(k)) < epsilon
        break
    end
    
    if (res(k+1) / res(k))>1
        if notfirst == 1
            E = E_old;
            Ah = Ah_old;
            ASh = ASh_old;

            break
        else
            notfirst = 1;
        end
    end
    
end
end

function [ E, A, res ] = lowResStep( H, E, A )

maxIter = 100;
epsilon = 1.01;
gamma1 = 1.01;

notfirst = 0;

res(1) = norm(H-E*A,'fro')+100;

for k=1:maxIter
    E_old = E;
    A_old = A;
    
    % 2.1. Update of signatures
    ck = gamma1 * norm(A*A','fro');
    U = E - 1/ck * ( E*A - H ) * A';
    E = Pplusa(U);
    
    % Calculation of residuals
%     res(k+1,1) = sqrt(norm(H-E*A,'fro')^2/size(H,1)/size(H,2));
    res(k+1,1) = norm(H-E*A,'fro');
    
    % Checks for exiting iteration
    if (1/res(k+1) * res(k)) < epsilon
        break
    end
    
    if (res(k+1) / res(k))>1
        if notfirst == 1
            E = E_old;
            A = A_old;

            break
        else
            notfirst = 1;
        end
    end
    
end
end

function U = Pplusa(U)
% max{0,U}
U(U<0) = 0;
U(U>1) = 1;
end

function [V, Vh] = PplusbAAh(V, Vh)
% Simplex Projection

V(V<0)     = 0;
Vh(Vh<0) = 0;



V_sum = sum(V, 1);
Vh_sum = sum(Vh, 1);

V_sum   = repmat(V_sum, size(V, 1), 1);
Vh_sum  = repmat(Vh_sum, size(Vh, 1), 1);

V  = V  ./ (V_sum + Vh_sum);
Vh = Vh ./ (V_sum + Vh_sum);


end








